
; Copied from various emacs lisp places...

(defconst conf-stringlit-re
  (concat
   ;; These fail if backslash-quote ends the string (not worth
   ;; fixing?).  They precede the short versions so that the first two
   ;; quotes don't look like an empty short string.
   ;;
   ;; long single quoted triple quoted strings (SQTQ),
   ;; with potential embedded single quotes
   "'''[^']*\\(\\('[^']\\|''[^']\\)[^']*\\)*'''"
   "\\|"
   ;; long double quoted triple quoted strings (DQTQ),
   ;; with potential embedded double quotes
   "\"\"\"[^\"]*\\(\\(\"[^\"]\\|\"\"[^\"]\\)[^\"]*\\)*\"\"\""
   "\\|"
   "'\\([^'\n\\]\\|\\\\.\\)*'"     ; single-quoted
   "\\|"                                ; or
   "\"\\([^\"\n\\]\\|\\\\.\\)*\""  ; double-quoted
   )
  "Regular expression matching a Conf string literal.")

(defconst conf-continued-re
  ;; This is tricky because a trailing backslash does not mean
  ;; continuation if it's in a comment
  (concat
   "\\(" "[^;#'\"\n\\]" "\\|" conf-stringlit-re "\\)*"
   "\\\\$")
  "Regular expression matching Conf backslash continuation lines.")

(defconst conf-blank-or-comment-re "\\s*\\($\\|#\\|;\\)"
  "Regular expression matching a blank or comment line.")

(defvar conf-mode-syntax-table nil "Syntax table in use in Conf-mode buffers.")
(if conf-mode-syntax-table ()
  (setq conf-mode-syntax-table (make-syntax-table))
  ; Both single and double quotes are "strings"
  (modify-syntax-entry ?\'  "\""  conf-mode-syntax-table)
  (modify-syntax-entry ?\"  "\""  conf-mode-syntax-table)

  ; Underscores are allowed as "symbol constituent"
  (modify-syntax-entry ?_  "_"  conf-mode-syntax-table)
  ; Backslash escapes; pound sign/semi collon starts comments that newlines end.
  (modify-syntax-entry ?\\ "\\" conf-mode-syntax-table)
  (modify-syntax-entry ?\# "<"  conf-mode-syntax-table)
  (modify-syntax-entry ?\; "<"  conf-mode-syntax-table)
  (modify-syntax-entry ?\n ">"  conf-mode-syntax-table)
  ; Square brackets act as parenthesis
  (modify-syntax-entry ?\[ "("  conf-mode-syntax-table)
  (modify-syntax-entry ?]  ")"  conf-mode-syntax-table))

(defconst conf-font-lock-keywords 'nil
  "Conf-specific syntax to be hilighted.")

; (define-derived-mode conf-mode lisp-mode "Conf"
(define-derived-mode conf-mode text-mode "Conf"
  "Major mode for editing Conf code.
This is a C mode variant customized for Conf's syntax.  It shares most of
C mode's features.  Turning on Conf mode runs `conf-mode-hook'."
  (set (make-local-variable 'paragraph-start) (concat page-delimiter "\\|$"))
  (set (make-local-variable 'paragraph-separate) paragraph-start)
  (set (make-local-variable 'comment-start) "\\s#\\s")
  (set (make-local-variable 'comment-start-skip) "")
  (setq font-lock-defaults '(nil nil nil)))
                                                                                
(provide 'conf-mode)
 
;; Set us up to load by default for .nas files
(setq auto-mode-alist (append '(("\\.conf$" . conf-mode))
                              auto-mode-alist))
